from pydantic import BaseModel, Field

from pangbank_api.models import Taxon, Taxonomy
from typing import Annotated

class PaginationParams(BaseModel):
    offset: int = 0
    limit: int = Field(default=20, le=100)


class FilterRelease(BaseModel):
    only_latest_release: bool | None = None


class FilterCollection(FilterRelease):
    collection_name: str | None = None
    collection_id: int | None = None


class FilterTaxon(BaseModel):
    taxon_name: Annotated[str | None, Field(min_length=3)] = None
    substring_taxon_match: bool = False


class FilterGenome(BaseModel):
    genome_name: str | None = None


class FilterPangenome(BaseModel):
    pangenome_name: str | None = None


class FilterGenomeMetadata(BaseModel):
    metadata_key: str | None = None
    metadata_value: str | None = None
    substring_match: bool = False


class FilterGenomeTaxon(FilterGenome, FilterTaxon):
    pass


class FilterCollectionTaxonGenome(
    FilterCollection,
    FilterTaxon,
    FilterGenome,
):
    pass


class FilterGenomeTaxonGenomePangenome(
    FilterCollection, FilterTaxon, FilterGenome, FilterPangenome
):
    pass


def get_taxonomies_from_taxa(taxa: list[Taxon]) -> list[Taxonomy]:
    taxonomy_source_to_taxonomy: dict[int, Taxonomy] = {}

    for taxon in taxa:
        if taxon.taxonomy_source_id in taxonomy_source_to_taxonomy:
            taxonomy = taxonomy_source_to_taxonomy[taxon.taxonomy_source_id]
            taxonomy.taxa.append(taxon)
        else:
            taxonomy = Taxonomy(taxa=[taxon], taxonomy_source=taxon.taxonomy_source)

            if taxon.taxonomy_source_id is not None:
                taxonomy_source_to_taxonomy[taxon.taxonomy_source_id] = taxonomy
            else:
                raise ValueError(
                    f"Taxon {taxon} has no taxonomy source associated with it"
                )
    taxonomies = list(taxonomy_source_to_taxonomy.values())
    return taxonomies



from datetime import datetime

from sqlalchemy import UniqueConstraint
from sqlmodel import Field, Relationship, SQLModel  # type: ignore


class GenomeTaxonLink(SQLModel, table=True):
    genome_id: int | None = Field(
        default=None, foreign_key="genome.id", primary_key=True
    )
    taxon_id: int | None = Field(default=None, foreign_key="taxon.id", primary_key=True)


class PangenomeTaxonLink(SQLModel, table=True):
    pangenome_id: int | None = Field(
        default=None, foreign_key="pangenome.id", primary_key=True
    )
    taxon_id: int | None = Field(default=None, foreign_key="taxon.id", primary_key=True)


class CollectionReleaseGenomeMetadataLink(SQLModel, table=True):
    collection_release_id: int | None = Field(
        default=None, foreign_key="collectionrelease.id", primary_key=True
    )
    genome_metadata_source_id: int | None = Field(
        default=None, foreign_key="genomemetadatasource.id", primary_key=True
    )


class GenomeInPangenomeMetric(SQLModel):
    model_config = {"populate_by_name": True}  # type: ignore

    # Metrics
    genome_name: str = Field(..., alias="Genome_name")
    contigs: int = Field(..., alias="Contigs")
    genes: int = Field(..., alias="Genes")
    fragmented_genes: int = Field(..., alias="Fragmented_genes")
    families: int = Field(..., alias="Families")
    families_with_fragments: int = Field(..., alias="Families_with_fragments")
    families_in_multicopy: int = Field(..., alias="Families_in_multicopy")
    soft_core_families: int = Field(..., alias="Soft_core_families")
    soft_core_genes: int = Field(..., alias="Soft_core_genes")
    exact_core_families: int = Field(..., alias="Exact_core_families")
    exact_core_genes: int = Field(..., alias="Exact_core_genes")
    persistent_genes: int = Field(..., alias="Persistent_genes")
    persistent_fragmented_genes: int = Field(..., alias="Persistent_fragmented_genes")
    persistent_families: int = Field(..., alias="Persistent_families")
    persistent_families_with_fragments: int = Field(
        ..., alias="Persistent_families_with_fragments"
    )
    persistent_families_in_multicopy: int = Field(
        ..., alias="Persistent_families_in_multicopy"
    )
    shell_genes: int = Field(..., alias="Shell_genes")
    shell_fragmented_genes: int = Field(..., alias="Shell_fragmented_genes")
    shell_families: int = Field(..., alias="Shell_families")
    shell_families_with_fragments: int = Field(
        ..., alias="Shell_families_with_fragments"
    )
    shell_families_in_multicopy: int = Field(..., alias="Shell_families_in_multicopy")
    cloud_genes: int = Field(..., alias="Cloud_genes")
    cloud_fragmented_genes: int = Field(..., alias="Cloud_fragmented_genes")
    cloud_families: int = Field(..., alias="Cloud_families")
    cloud_families_with_fragments: int = Field(
        ..., alias="Cloud_families_with_fragments"
    )
    cloud_families_in_multicopy: int = Field(..., alias="Cloud_families_in_multicopy")
    completeness: float = Field(..., alias="Completeness")
    contamination: float = Field(..., alias="Contamination")
    fragmentation: float = Field(..., alias="Fragmentation")
    rgps: int = Field(..., alias="RGPs")
    spots: int = Field(..., alias="Spots")
    modules: int = Field(..., alias="Modules")

    # info extracted from metadata
    strain: str | None = None
    organism_name: str | None = None

class GenomePangenomeLink(GenomeInPangenomeMetric, table=True):
    __table_args__ = (UniqueConstraint("genome_id", "pangenome_id"),)

    id: int | None = Field(default=None, primary_key=True)
    genome_id: int | None = Field(default=None, foreign_key="genome.id")
    pangenome_id: int | None = Field(default=None, foreign_key="pangenome.id")

    pangenome: "Pangenome" = Relationship(back_populates="genome_links")
    genome: "Genome" = Relationship(back_populates="pangenome_links")

    genome_file_md5sum: str
    genome_file_name: str


class CollectionBase(SQLModel):
    name: str = Field(unique=True)
    description: str | None = None


class Collection(CollectionBase, table=True):
    id: int | None = Field(default=None, primary_key=True)

    releases: list["CollectionRelease"] = Relationship(
        back_populates="collection", cascade_delete=True
    )


class CollectionPublic(CollectionBase):
    id: int


class CollectionPublicWithReleases(CollectionPublic):
    releases: list["CollectionReleasePublicWithCount"]


class TaxonomySourceBase(SQLModel):
    name: str
    ranks: str
    version: str | None = None
    description: str | None = None
    source: str | None = None
    url: str | None = None


class TaxonomySource(TaxonomySourceBase, table=True):
    __table_args__ = (UniqueConstraint("name", "version", name="uq_name_version"),)

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field()

    collection_releases: list["CollectionRelease"] = Relationship(
        back_populates="taxonomy_source"
    )
    taxa: list["Taxon"] = Relationship(back_populates="taxonomy_source")


class TaxonomySourcePublic(TaxonomySourceBase):
    id: int

    def __str__(self):
        return " ".join([self.name, self.version or ""])


class GenomeSourceBase(SQLModel):
    name: str
    version: str | None = None
    description: str | None = None
    source: str | None = None
    url: str | None = None


class GenomeSource(GenomeSourceBase, table=True):
    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(unique=True)  # Ensuring name uniqueness at the model level

    genomes: list["Genome"] = Relationship(back_populates="genome_source")


class GenomeSourcePublic(GenomeSourceBase):
    id: int


class CollectionReleaseBase(SQLModel):
    version: str

    ppanggolin_version: str
    pangbank_wf_version: str

    release_note: str

    mash_version: str
    mash_sketch_md5sum: str

    latest: bool = False
    date: datetime
    collection_id: int | None = Field(
        default=None, foreign_key="collection.id", ondelete="CASCADE"
    )

    # RESTRICT: Prevent the deletion of this record (CollectionReleaseBase)
    # if there is a foreign key value by raising an error.
    taxonomy_source_id: int | None = Field(
        default=None, foreign_key="taxonomysource.id", ondelete="RESTRICT"
    )


class CollectionRelease(CollectionReleaseBase, table=True):
    id: int | None = Field(default=None, primary_key=True)

    mash_sketch: str

    pangenomes_directory: str
    collection: "Collection" = Relationship(back_populates="releases")

    taxonomy_source: "TaxonomySource" = Relationship(
        back_populates="collection_releases"
    )

    pangenomes: list["Pangenome"] = Relationship(
        back_populates="collection_release", cascade_delete=True
    )

    genome_metadata_sources: list["GenomeMetadataSource"] = Relationship(
        back_populates="collection_releases",
        link_model=CollectionReleaseGenomeMetadataLink,
    )

class CollectionReleasePublic(CollectionReleaseBase):
    id: int
    taxonomy_source: TaxonomySourcePublic
    collection_name: str
    collection: CollectionPublic

class CollectionReleasePublicWithCount(CollectionReleasePublic):
    pangenome_count: int


class TaxonBase(SQLModel):
    name: str = Field(index=True)
    rank: str
    depth: int

    taxid: int | None = None

    taxonomy_source_id: int | None = Field(
        index=True, default=None, foreign_key="taxonomysource.id"
    )

    def __lt__(self, other: "Taxon"):
        return self.name < other.name


class Taxon(TaxonBase, table=True):
    id: int | None = Field(default=None, primary_key=True)

    taxonomy_source: TaxonomySource = Relationship(back_populates="taxa")

    genomes: list["Genome"] = Relationship(
        back_populates="taxa", link_model=GenomeTaxonLink
    )
    pangenomes: list["Pangenome"] = Relationship(
        back_populates="taxa", link_model=PangenomeTaxonLink
    )


class TaxonPublic(TaxonBase):
    id: int


class GenomeBase(SQLModel):
    name: str = Field(unique=True, index=True)
    genome_source_id: int | None = Field(default=None, foreign_key="genomesource.id")


class Genome(GenomeBase, table=True):
    id: int | None = Field(default=None, primary_key=True)
    # version : str | None = None
    # collection_releases : list[CollectionRelease] = Relationship(back_populates="genomes",
    # link_model=GenomeCollectionReleaseLink)

    taxa: list[Taxon] = Relationship(
        back_populates="genomes", link_model=GenomeTaxonLink
    )

    pangenome_links: list[GenomePangenomeLink] = Relationship(back_populates="genome")

    genome_source: GenomeSource = Relationship(back_populates="genomes")

    genome_metadata: list["GenomeMetadata"] = Relationship(
        back_populates="genome", cascade_delete=True
    )


class GenomePublic(GenomeBase):
    id: int
    genome_source: GenomeSourcePublic


class GenomePublicWithTaxonomies(GenomePublic):
    taxonomies: list["TaxonomyPublic"]


class PangenomeMetric(SQLModel):
    # Metrics
    gene_count: int
    genome_count: int
    family_count: int
    edge_count: int

    # Persistent information
    persistent_family_count: int
    persistent_family_min_genome_frequency: float
    persistent_family_max_genome_frequency: float
    persistent_family_std_genome_frequency: float
    persistent_family_mean_genome_frequency: float

    # Shell information
    shell_family_count: int
    shell_family_min_genome_frequency: float
    shell_family_max_genome_frequency: float
    shell_family_std_genome_frequency: float
    shell_family_mean_genome_frequency: float

    # Cloud information
    cloud_family_count: int
    cloud_family_min_genome_frequency: float
    cloud_family_max_genome_frequency: float
    cloud_family_std_genome_frequency: float
    cloud_family_mean_genome_frequency: float

    partition_count: int
    rgp_count: int
    spot_count: int

    all_genome_fluidity: float | None = None
    shell_genome_fluidity: float | None = None
    cloud_genome_fluidity: float | None = None
    # Accessory genome fluidity of shell and cloud genome fluidity
    accessory_genome_fluidity: float | None = None

    # Modules information
    module_count: int
    family_in_module_count: int

    # Summary metrics from genomes
    mean_completeness: float
    mean_contamination: float
    mean_fragmentation: float

    mean_exact_core_families_count_per_genome: float
    mean_soft_core_families_count_per_genome: float
    mean_persistent_families_count_per_genome: float
    mean_shell_families_count_per_genome: float
    mean_cloud_families_count_per_genome: float


class PangenomeBase(PangenomeMetric):

    annotation_source: str | None = None
    name: str
    file_md5sum: str


class Pangenome(PangenomeBase, table=True):
    id: int | None = Field(default=None, primary_key=True)
    name: str
    file_name: str

    collection_release_id: int | None = Field(
        default=None, foreign_key="collectionrelease.id"
    )

    collection_release: CollectionRelease = Relationship(back_populates="pangenomes")

    genome_links: list[GenomePangenomeLink] = Relationship(
        back_populates="pangenome", cascade_delete=True
    )

    taxa: list[Taxon] = Relationship(
        back_populates="pangenomes", link_model=PangenomeTaxonLink
    )


class PangenomePublic(PangenomeBase):
    id: int

    collection_release: CollectionReleasePublic

    taxonomy: "TaxonomyPublic"

    genome_count: int

    collection_release_id: int

    persistent_fraction: float
    shell_fraction: float
    cloud_fraction: float

class TaxonomyBase(SQLModel):
    pass


class Taxonomy(TaxonomyBase):
    taxonomy_source: TaxonomySource
    taxa: list[Taxon]


class TaxonomyPublic(TaxonomyBase):
    taxonomy_source: TaxonomySourcePublic
    taxa: list[TaxonPublic]

    def __str__(self):
        return ";".join([taxon.name for taxon in self.taxa])

    def __lt__(self, other: "TaxonomyPublic"):
        return self.taxa < other.taxa


class MetadataBase(SQLModel):

    key: str = Field(index=True)
    value: str


class GenomeMetadata(MetadataBase, table=True):
    id: int | None = Field(default=None, primary_key=True)

    source_id: int | None = Field(
        foreign_key="genomemetadatasource.id", default=None, ondelete="CASCADE"
    )

    genome_id: int | None = Field(
        foreign_key="genome.id", default=None, ondelete="CASCADE"
    )
    genome: Genome = Relationship(back_populates="genome_metadata")

    source: "GenomeMetadataSource" = Relationship(back_populates="genome_metadata")


class MetadataSourceBase(SQLModel):

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(unique=True)
    description: str | None = None
    url: str | None = None
    strain_attribute: str | None = None
    organism_name_attribute: str | None = None


class GenomeMetadataSource(MetadataSourceBase, table=True):
    genome_metadata: list[GenomeMetadata] = Relationship(
        back_populates="source", cascade_delete=True
    )
    collection_releases: list[CollectionRelease] = Relationship(
        back_populates="genome_metadata_sources",
        link_model=CollectionReleaseGenomeMetadataLink,
    )


class GenomePangenomeLinkPublic(GenomeInPangenomeMetric):
    genome_id: int
    pangenome_id: int

    genome_file_md5sum: str
    genome_file_name: str


class GenomePangenomeLinkWithMetadataPublic(GenomePangenomeLinkPublic):
    genome_metadata: list[MetadataBase]



from pydantic_settings import BaseSettings
from functools import lru_cache
from pathlib import Path


class Settings(BaseSettings):
    pangbank_db_path: Path = Path("database/database.db")
    pangbank_data_dir: Path = Path("data/")
    pangbank_origins: str = (
        "http://localhost:3000"  # list of origins separated by semicolon
    )


@lru_cache
def get_settings():
    return Settings()



from pathlib import Path
from typing import Iterator, Sequence

from sqlalchemy import func

from sqlalchemy.orm import aliased

from sqlmodel import Session, select

from pangbank_api.crud.common import (
    FilterGenomeTaxonGenomePangenome,
    PaginationParams,
    FilterGenome,
    FilterGenomeMetadata,
    get_taxonomies_from_taxa,
)
from pangbank_api.models import (
    Genome,
    GenomePangenomeLink,
    Pangenome,
    PangenomePublic,
    PangenomeTaxonLink,
    Taxon,
    CollectionRelease,
    CollectionReleasePublic,
    Collection,
    TaxonomySourcePublic,
    CollectionPublic,
    TaxonomyPublic,
)


def get_pangenome_file(session: Session, pangenome_id: int) -> Path | None:
    pangenome = session.get(Pangenome, pangenome_id)

    if not pangenome:
        return None

    pangenome_file = (
        Path(pangenome.collection_release.pangenomes_directory) / pangenome.file_name
    )

    return pangenome_file


def get_pangenome(session: Session, pangenome_id: int) -> Pangenome | None:
    pangenome = session.get(Pangenome, pangenome_id)
    if pangenome is None:
        return None

    return pangenome


def make_pangenome_public_metrics(p: Pangenome) -> dict[str, float]:
    mean_fam_per_genome = p.mean_persistent_families_count_per_genome + p.mean_shell_families_count_per_genome + p.mean_cloud_families_count_per_genome
    return {
        "persistent_fraction": p.mean_persistent_families_count_per_genome / mean_fam_per_genome,
        "shell_fraction": p.mean_shell_families_count_per_genome / mean_fam_per_genome,
        "cloud_fraction": p.mean_cloud_families_count_per_genome / mean_fam_per_genome,
    }


def make_pangenome_public(pangenome: Pangenome) -> PangenomePublic:

    taxonomies = get_taxonomies_from_taxa(pangenome.taxa)

    assert (
        len(taxonomies) == 1
    ), f"{pangenome.file_name} {pangenome.id} have an issue with its taxonomies. Found {len(taxonomies)} taxonomies"

    collection_release_public = CollectionReleasePublic(
        **pangenome.collection_release.model_dump(),
        taxonomy_source=TaxonomySourcePublic(
            **pangenome.collection_release.taxonomy_source.model_dump()
        ),
        collection_name=pangenome.collection_release.collection.name,
        collection=CollectionPublic(
            **pangenome.collection_release.collection.model_dump()
        ),
    )

    pangenome_public = PangenomePublic(
        **pangenome.model_dump(),  # type: ignore
        **make_pangenome_public_metrics(pangenome),
        collection_release=collection_release_public,
        taxonomy=TaxonomyPublic(**taxonomies[0].model_dump()),
    )

    return pangenome_public


def get_public_pangenome(session: Session, pangenome_id: int) -> PangenomePublic | None:

    pangenome = get_pangenome(session, pangenome_id)
    if pangenome is None:
        return None

    return make_pangenome_public(pangenome)


def get_pangenomes(
    session: Session,
    filter_params: FilterGenomeTaxonGenomePangenome | None = None,
    pagination_params: PaginationParams | None = None,
) -> Sequence[Pangenome]:

    query = select(Pangenome).distinct()

    if filter_params and filter_params.pangenome_name is not None:
        query = query.where(Pangenome.name == filter_params.pangenome_name)

    collectionrelease_alias = aliased(CollectionRelease)
    # collectionrelease_alias_2 = aliased(CollectionRelease)

    if filter_params and filter_params.only_latest_release is True:
        query = query.join(collectionrelease_alias).where(
            collectionrelease_alias.latest
        )

    if filter_params and filter_params.collection_name is not None:
        query = (
            query.join(CollectionRelease)
            .join(Collection)
            .where(Collection.name == filter_params.collection_name)
        )

    if filter_params and filter_params.collection_id is not None:
        query = (
            query.join(CollectionRelease)
            .join(Collection)
            .where(Collection.id == filter_params.collection_id)
        )

    if filter_params and filter_params.genome_name is not None:
        query = (
            query.join(GenomePangenomeLink)
            .join(Genome)
            .where(Genome.name == filter_params.genome_name)
        )

    if filter_params and filter_params.taxon_name is not None:
        # Apply offset and limit

        query = query.join(PangenomeTaxonLink).join(Taxon)
        if filter_params.substring_taxon_match:
            query = query.where(
                func.lower(Taxon.name).like(f"%{filter_params.taxon_name.lower()}%")
            )
        else:
            # exact match
            query = query.where(Taxon.name == filter_params.taxon_name)
    # Apply offset and limit
    if pagination_params:
        query = query.offset(pagination_params.offset).limit(pagination_params.limit)

    pangenomes = session.exec(query).all()

    return pangenomes


def get_public_pangenomes(
    session: Session,
    filter_params: FilterGenomeTaxonGenomePangenome | None = None,
    pagination_params: PaginationParams | None = None,
) -> Iterator[PangenomePublic]:

    pangenomes = get_pangenomes(session, filter_params, pagination_params)

    public_pangenomes = (make_pangenome_public(pangenome) for pangenome in pangenomes)

    return public_pangenomes


def get_genomes_in_pangenome(
    session: Session,
    pangenome_id: int,
    filter_genome: FilterGenome | None = None,
    filter_metadata: FilterGenomeMetadata | None = None,
    pagination_params: PaginationParams | None = None,
):
    # Alias for the metadata table
    # metadata_alias = aliased(GenomeInPangenomeMetadata)

    query = (
        select(GenomePangenomeLink)
        .distinct()
        .join(Pangenome)
        .where(Pangenome.id == pangenome_id)
    )

    if filter_genome and filter_genome.genome_name is not None:
        query = query.join(Genome).where(Genome.name == filter_genome.genome_name)

    if pagination_params:
        query = query.offset(pagination_params.offset).limit(pagination_params.limit)

    # if filter_metadata:
    #     if filter_metadata.metadata_key:
    #         query = query.join(metadata_alias).where(
    #             metadata_alias.key == filter_metadata.metadata_key
    #         )

    #     if filter_metadata.metadata_value is not None:
    #         query = query.where(metadata_alias.value == filter_metadata.metadata_value)

    pangenome_genomes_links = session.exec(query).all()
    return list(pangenome_genomes_links)


def get_genome_in_pangenome(session: Session, pangenome_id: int, genome_id: int):

    query = (
        select(GenomePangenomeLink)
        .distinct()
        .where(
            (GenomePangenomeLink.pangenome_id == pangenome_id)
            & (GenomePangenomeLink.genome_id == genome_id)
        )
    )

    pangenome_genome_link = session.exec(query).first()
    return pangenome_genome_link



from fastapi import APIRouter, Depends, HTTPException

from pangbank_api.crud import genomes as genomes_crud
from pangbank_api.crud.common import FilterGenomeTaxon, PaginationParams
from pangbank_api.dependencies import SessionDep
from pangbank_api.models import GenomePublicWithTaxonomies

router = APIRouter(
    tags=["genomes"],
    responses={404: {"description": "Not found"}},
)


@router.get("/genomes/", response_model=list[GenomePublicWithTaxonomies])
async def read_genomes(
    session: SessionDep,
    filter_params: FilterGenomeTaxon = Depends(),
    pagination_params: PaginationParams = Depends(),
):
    genomes = genomes_crud.get_genomes(session, filter_params, pagination_params)

    return genomes


@router.get("/genomes/{genome_id}", response_model=GenomePublicWithTaxonomies)
async def get_genome_by_id(genome_id: int, session: SessionDep):
    # genome = session.get(Genome, genome_id)
    genome = genomes_crud.get_genome_by_id(session=session, genome_id=genome_id)
    if not genome:
        raise HTTPException(status_code=404, detail="Genome not found")
    return genome



from typing import List, Sequence

from pathlib import Path

from sqlmodel import Session, select
from packaging.version import parse

from pangbank_api.crud.common import FilterCollection, FilterRelease
from pangbank_api.models import (
    Collection,
    CollectionReleasePublicWithCount,
    CollectionPublicWithReleases,
)


def get_collections(
    session: Session, filter_params: FilterCollection
) -> Sequence[CollectionPublicWithReleases]:
    query = select(Collection)

    # Check if filter_params.collection_release_id is provided
    if filter_params.collection_name is not None:
        query = query.where(Collection.name == filter_params.collection_name)

    collections = session.exec(query).all()

    public_collections: List[CollectionPublicWithReleases] = []

    for collection in collections:
        collection_public = make_collection_public_with_releases(
            collection, filter_params.only_latest_release
        )
        public_collections.append(collection_public)

    return public_collections


def make_collection_public_with_releases(
    collection: Collection, only_latest_release: bool | None
):

    public_releases: List[CollectionReleasePublicWithCount] = []

    latest_release = max(collection.releases, key=lambda x: parse(x.version))
    assert (
        latest_release.latest
    ), f"Latest release {latest_release} is not marked as latest"

    if only_latest_release:
        releases = [latest_release]
    else:
        releases = collection.releases

    for release in releases:

        release_public = CollectionReleasePublicWithCount.model_validate(
            release,
            from_attributes=True,
            update={
                "pangenome_count": len(release.pangenomes),
                "collection_name": collection.name,
            },
        )
        public_releases.append(release_public)

    # Sort using semantic versioning
    public_releases = sorted(
        public_releases, key=lambda x: parse(x.version), reverse=True
    )

    collection_public = CollectionPublicWithReleases.model_validate(
        collection,
        from_attributes=True,
        update={
            "releases": public_releases,
        },
    )
    return collection_public


def get_collection(session: Session, collection_id: int, filter_release: FilterRelease):

    collection = session.get(Collection, collection_id)
    if collection is None:
        return None

    public_collection = make_collection_public_with_releases(
        collection, only_latest_release=filter_release.only_latest_release
    )
    return public_collection


def get_collection_mash_sketch(session: Session, collection_id: int):

    collection = session.get(Collection, collection_id)

    if not collection:
        return None

    latest_release = max(collection.releases, key=lambda x: parse(x.version))

    return Path(latest_release.mash_sketch)

def get_collection_index_directory(session: Session, collection_id: int):

    collection = session.get(Collection, collection_id)

    if not collection:
        return None

    latest_release = max(collection.releases, key=lambda x: parse(x.version))

    return Path(latest_release.pangenomes_directory) / ".." / "metapang/bank_index"


import csv
import gzip
import logging
from pathlib import Path
from typing import Generator, List, Optional
from rich.console import Console
from rich.table import Table

import typer
from rich.progress import track
from sqlmodel import Session, select
from typing_extensions import Annotated

from pangbank_api.database import create_db_and_tables, engine
from pangbank_api.manage_db.utils import check_and_read_json_file, set_up_logging_config
from pangbank_api.models import (
    Genome,
    GenomeMetadata,
    MetadataBase,
    GenomeMetadataSource,
    GenomePangenomeLink,
    CollectionRelease,
    Pangenome,
)
from sqlalchemy import exists
from sqlalchemy.orm import selectinload

logger = logging.getLogger(__name__)  # __name__ ensures uniqueness per module

app = typer.Typer(no_args_is_help=True)


def parse_metadata_table(
    file_path: Path,
    disable_track: bool = False,
) -> Generator[tuple[str, List[MetadataBase]], None, None]:
    """Parse a gzip-compressed TSV and yield rows as dictionaries."""
    proper_open = gzip.open if file_path.name.endswith("gz") else open
    with proper_open(file_path, mode="rt") as tsvfile:
        reader = csv.DictReader(tsvfile, delimiter="\t")
        for row in track(reader, "Parsing metadata", total=None, disable=disable_track):
            try:
                genome_name = row["genomes"]
            except KeyError:
                raise KeyError(
                    f"Missing 'genomes' column in {file_path}. "
                    "Ensure the TSV contains a 'genomes' column."
                )
            genome_metadata_list = [
                MetadataBase(
                    key=key,
                    value=str(value),
                )
                for key, value in row.items()
                if key != "genomes"
            ]

            yield genome_name, genome_metadata_list


def get_genomes_from_release_without_metadata_in_batches(
    collection_release: CollectionRelease,
    metadata_source: GenomeMetadataSource,
    session: Session,
    chunk_size: int = 1000,
):
    offset = 0

    while True:
        metadata_exists_subquery = select(GenomeMetadata.id).where(
            GenomeMetadata.genome_id == Genome.id,
            GenomeMetadata.source_id == metadata_source.id,
        )

        genomes_statement = (
            select(Genome)
            .join(GenomePangenomeLink)
            .join(Pangenome)
            .where(Pangenome.collection_release_id == collection_release.id)
            .where(~exists(metadata_exists_subquery))
            .distinct()
            .offset(offset)
            .limit(chunk_size)
        )

        genomes_batch = session.exec(genomes_statement).all()

        if not genomes_batch:
            break

        yield genomes_batch
        offset += chunk_size


def add_metadata_to_genomes_of_the_release(
    collection_release: CollectionRelease,
    metadata_source_and_genome_name_to_metadatas: list[
        tuple[GenomeMetadataSource, dict[str, list[MetadataBase]]]
    ],
    session: Session,
):

    for (
        metadata_source,
        genome_name_to_metadatas,
    ) in metadata_source_and_genome_name_to_metadatas:

        logger.info(
            f"Adding metadata to genomes of the release from metadata source {metadata_source.name}."
        )

        for genomes in get_genomes_from_release_without_metadata_in_batches(
            collection_release, metadata_source, session
        ):

            logger.info(
                f"Retrieved {len(genomes)} genomes from the database "
                f"that are part of collection release {collection_release.id} "
                f"and that have no metadata from source {metadata_source.name} ({metadata_source.description})."
            )

            all_metadatas: List[GenomeMetadata] = []

            for genome in genomes:
                genome_name = genome.name
                if genome.id is None:
                    raise ValueError(
                        f"Genome {genome_name} does not have an ID. Cannot add metadata."
                    )
                if genome_name in genome_name_to_metadatas:
                    metadatabase_list = genome_name_to_metadatas[genome_name]
                    all_metadatas += create_metadata(
                        genome.id, metadatabase_list, metadata_source
                    )

            logger.info(
                f"Adding {len(all_metadatas)} new metadata from metadata source {metadata_source.name} to the database describing {len(genomes)} genomes."
            )
            session.add_all(all_metadatas)

    session.commit()


def get_all_genome_pangenome_links_for_release(
    collection_release: CollectionRelease,
    session: Session,
):
    statement = (
        select(GenomePangenomeLink)
        .options(selectinload(GenomePangenomeLink.genome))  # type: ignore
        .join(Pangenome)
        .where(Pangenome.collection_release_id == collection_release.id)
    )
    results = session.exec(statement).all()

    return results


def update_genome_pangenome_links_with_specific_metadata(
    collection_release: CollectionRelease,
    metadata_source_and_genome_name_to_metadatas: list[
        tuple[GenomeMetadataSource, dict[str, list[MetadataBase]]]
    ],
    session: Session,
):

    genome_pangenome_links = get_all_genome_pangenome_links_for_release(
        collection_release, session
    )
    genome_name_to_genome_pangenome_link = {
        genome_pangenome_link.genome.name: genome_pangenome_link
        for genome_pangenome_link in genome_pangenome_links
    }

    updated_genome_pangenome_links: List[GenomePangenomeLink] = []

    for (
        genome_metadata_source,
        genome_name_to_metadatas,
    ) in metadata_source_and_genome_name_to_metadatas:

        for genome_name, metadatas in genome_name_to_metadatas.items():
            if genome_name in genome_name_to_genome_pangenome_link:
                is_updated = False
                genome_pangenome_link = genome_name_to_genome_pangenome_link[
                    genome_name
                ]

                strain_attribute = genome_metadata_source.strain_attribute
                organism_name_attribute = genome_metadata_source.organism_name_attribute

                for metadata in metadatas:

                    if metadata.key == strain_attribute:
                        strain = metadata.value
                        genome_pangenome_link.strain = strain
                        is_updated = True

                    if metadata.key == organism_name_attribute:
                        organism_name = metadata.value
                        genome_pangenome_link.organism_name = organism_name
                        is_updated = True

                if is_updated:
                    updated_genome_pangenome_links.append(genome_pangenome_link)

        logger.info(
            f"Updating {len(updated_genome_pangenome_links)} genome pangenome links with strain and organism name from source {genome_metadata_source.name}."
        )

        session.add_all(updated_genome_pangenome_links)

        session.commit()


def create_metadata(
    genome_id: int,
    metadata_list: List[MetadataBase],
    source: GenomeMetadataSource,
):
    """ """
    metadatas: List[GenomeMetadata] = []

    for metadata_input in metadata_list:
        metadata = GenomeMetadata.model_validate(
            metadata_input, update={"genome_id": genome_id, "source_id": source.id}
        )
        metadatas.append(metadata)

    return metadatas


def parse_metadata_source_file(metadata_source_file: Path):
    json_content = check_and_read_json_file(metadata_source_file)

    metadata_source = GenomeMetadataSource.model_validate(json_content)

    return metadata_source


def get_all_genomes_in_pangenome(session: Session):
    genomes_statement = (
        select(Genome)
        .join(GenomePangenomeLink)
        .where(Genome.id == GenomePangenomeLink.genome_id)
    )
    genomes_with_links = session.exec(genomes_statement).all()

    return genomes_with_links


def add_genome_metadata_source_to_db(
    metadata_source: GenomeMetadataSource, session: Session
):

    metadata_source_in_db = session.exec(
        select(GenomeMetadataSource).where(
            (GenomeMetadataSource.name == metadata_source.name)
        )
    ).first()

    if metadata_source_in_db is not None:
        logger.info(
            f"Genome Metadata Source '{metadata_source.name}' already exists in the database."
        )
        return metadata_source_in_db

    else:
        logger.info(
            f"Adding new Genome Metadata Source '{metadata_source.name}' to the database."
        )
        session.add(metadata_source)
        session.commit()
        session.refresh(metadata_source)
        return metadata_source


@app.command(no_args_is_help=True)
def add(
    metadata_source_file: Annotated[
        Path,
        typer.Argument(
            help="Path to a JSON file that describes the metadata source"
            " contains file information with genome metadata. "
        ),
    ],
    metadata_file: Annotated[
        Path,
        typer.Argument(
            help="Path to a TSV file that contains genome metadata. A column 'Genome' must be present."
        ),
    ],
):
    set_up_logging_config()

    metadata_source = parse_metadata_source_file(metadata_source_file)

    if not metadata_file.exists():
        typer.echo(
            f"Error: The file {metadata_file} does not exist",
            err=True,
        )
        raise typer.Exit(1)

    create_db_and_tables()

    with Session(engine) as session:
        add_genome_metadata_source_to_db(metadata_source, session)

        # we could add more filter to add metadata to only genomes of interest
        genomes = get_all_genomes_in_pangenome(session)

        genome_name_to_genome_id = {genome.name: genome.id for genome in genomes}

        logging.info(
            f"Retrieved {len(genome_name_to_genome_id)} genomes within a pangenome from the database."
        )

        metadada_list: List[GenomeMetadata] = []

        unknown_genome_count = 0
        genome_processed_count = 0
        logging.info(f"Parsing genome metadata from {metadata_file}.")

        genome_to_metadata = [
            (genome, metdata) for genome, metdata in parse_metadata_table(metadata_file)
        ]

        logging.info(
            f"Metadata for {len(genome_to_metadata)} genomes have been collected."
        )
        for genome_name, genome_metadata in track(
            genome_to_metadata,
            description="Processing metadata",
            total=len(genome_to_metadata),
        ):
            genome_id = genome_name_to_genome_id.get(genome_name)

            if genome_id is not None:
                metadada_list += create_metadata(
                    genome_id, genome_metadata, metadata_source
                )
                genome_processed_count += 1
            else:
                unknown_genome_count += 1

            if genome_processed_count == 10000:
                logging.info(
                    f"Adding new {len(metadada_list)} metadata to the database describing {genome_processed_count} genomes."
                )

                session.add_all(metadada_list)
                session.commit()
                genome_processed_count = 0
                metadada_list = []

        logging.info(
            f"Adding final {len(metadada_list)} metadata to the database describing {genome_processed_count} genomes."
        )
        session.add_all(metadada_list)

        if unknown_genome_count > 0:
            logging.info(
                f"{unknown_genome_count} genomes in the metadata file are not in the database."
            )

        session.commit()


@app.command(no_args_is_help=True)
def delete(
    metadata_source_name: Annotated[
        str,
        typer.Argument(help="Name of the metadata source to delete from the database."),
    ],
    metadata_source_version: Annotated[
        Optional[str],
        typer.Option(
            help="Specific metadata source version to delete. If not provided, the entire source will be deleted."
        ),
    ] = None,
):
    """
    Deletes a collection from the database if it exists.

    :param session: Database session used for querying and deleting the collection.
    :param collection_name: Name of the collection to delete.
    """

    set_up_logging_config()

    create_db_and_tables()

    with Session(engine) as session:
        if metadata_source_version is None:
            # Query the database to find the collection with the specified name
            statement = select(GenomeMetadataSource).where(
                GenomeMetadataSource.name == metadata_source_name
            )

        else:
            statement = select(GenomeMetadataSource).where(
                (GenomeMetadataSource.name == metadata_source_name)
            )

        metadata_sources = session.exec(statement).all()

        if not metadata_sources:
            source_info = (
                f"{metadata_source_name} version: {metadata_source_version}"
                if metadata_source_version
                else metadata_source_name
            )
            error_message = f"Genome Metadata Source '{source_info}' not found in the database. Deletion aborted. "

            metadata_sources = session.exec(select(GenomeMetadataSource)).all()

            avalaible_metadata = [
                f"name='{source.name}'" for source in metadata_sources
            ]

            error_message += f"Available genome metadata sources in the database: {avalaible_metadata}"
            raise ValueError(error_message)

        else:
            for source in metadata_sources:
                logging.info(
                    f"Deleting genome metadata source: '{source.name}' from the database."
                )
                session.delete(source)

            session.commit()


@app.command()
def list():
    """
    List all genome metadata sources in the database.
    """

    set_up_logging_config()
    console = Console()

    with Session(engine) as session:
        metadata_sources = session.exec(select(GenomeMetadataSource)).all()

        if not metadata_sources:
            console.print(
                "[bold red]No genome metadata sources found in the database.[/bold red]"
            )
            return

        # Create a table for the metadata sources
        table = Table(title="Genome Metadata Sources", header_style="bold magenta")
        table.add_column("Name", style="bold cyan")

        for source in metadata_sources:
            table.add_row(source.name)

        console.print(table)



import json
import logging
from pathlib import Path
import hashlib
from sqlmodel import SQLModel  # type: ignore

import typer
from rich.logging import RichHandler

from pangbank_api.manage_db.input_models import CollectionReleaseInput


def check_and_read_json_file(input_json_file: Path):
    if not input_json_file.exists():
        typer.echo(
            f"[bold red]Error:[/bold red] JSON file '{input_json_file}' does not exist.",
            err=True,
        )
        raise typer.Exit(1)

    try:
        with open(input_json_file) as fl:
            json_content = json.load(fl)

    except json.JSONDecodeError as e:
        typer.echo(f"[bold red]Error:[/bold red] Invalid JSON format: {e}", err=True)
        raise typer.Exit(1)

    return json_content


def parse_collection_release_input_json(
    collection_release_json: Path, pangbank_data_dir: Path
):
    json_content = check_and_read_json_file(collection_release_json)
    mash_sketch_md5sum = compute_md5(
        pangbank_data_dir / json_content["release"]["mash_sketch"]
    )
    json_content["release"]["mash_sketch_md5sum"] = mash_sketch_md5sum
    # Validate JSON structure using Pydantic
    data_input = CollectionReleaseInput.model_validate(json_content)

    pangenomes_directory = pangbank_data_dir / data_input.release.pangenomes_directory
    mash_sketch_file = pangbank_data_dir / data_input.release.mash_sketch

    # Check if paths exist
    missing_files = (
        [
            data_input.taxonomy.file,
            pangenomes_directory,
            mash_sketch_file,
        ]
        + [gs.file for gs in data_input.genome_sources]
        + [
            metadata_source.file
            for metadata_source in data_input.genome_metadata_sources
        ]
    )

    missing_files = [f for f in missing_files if not f.exists()]

    if missing_files:
        typer.echo(
            "[bold red]Error:[/bold red] The following files are missing:", err=True
        )
        for f in missing_files:
            typer.echo(f"  - {f}", err=True)
        raise typer.Exit(1)

    return data_input


def set_up_logging_config():
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(levelname)s - %(message)s",
        handlers=[RichHandler()],
    )


def compute_md5(file_path: Path):
    md5_hash = hashlib.md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):  # Read file in chunks
            md5_hash.update(chunk)
    return md5_hash.hexdigest()



from fastapi import APIRouter, Depends, HTTPException
from fastapi.responses import FileResponse

from pangbank_api.crud import pangenomes as pangenomes_crud
from pangbank_api.crud.common import (
    FilterGenomeTaxonGenomePangenome,
    PaginationParams,
    FilterGenome,
)

from pangbank_api.dependencies import SessionDep
from pangbank_api.models import PangenomePublic, GenomePangenomeLinkPublic
from pangbank_api.dependencies import SettingsDep
from pathlib import Path

router = APIRouter(
    tags=["pangenomes"],
)


@router.get("/pangenomes/", response_model=list[PangenomePublic])
async def get_pangenomes(
    session: SessionDep,
    filter_params: FilterGenomeTaxonGenomePangenome = Depends(),
    pagination_params: PaginationParams = Depends(),
):
    pangenomes = pangenomes_crud.get_public_pangenomes(
        session, filter_params, pagination_params
    )

    return list(pangenomes)


@router.get("/pangenomes/{pangenome_id}", response_model=PangenomePublic)
async def get_pangenome(pangenome_id: int, session: SessionDep):
    pangenome = pangenomes_crud.get_public_pangenome(session, pangenome_id)

    if not pangenome:
        raise HTTPException(status_code=404, detail="Pangenome not found")
    return pangenome


@router.get(
    "/pangenomes/{pangenome_id}/file", response_model=str, response_class=FileResponse
)
async def get_pangenome_file(
    pangenome_id: int, session: SessionDep, settings: SettingsDep
):
    pangenome = pangenomes_crud.get_pangenome(session, pangenome_id)

    if not pangenome:
        raise HTTPException(status_code=404, detail="Pangenome not found")

    pangenome_relative_path = (
        Path(pangenome.collection_release.pangenomes_directory) / pangenome.file_name
    )

    pangenome_full_path = settings.pangbank_data_dir / pangenome_relative_path
    if not pangenome_full_path.exists():
        raise HTTPException(
            status_code=404,
            detail=f"Pangenome file {pangenome_relative_path} does not exists",
        )

    return FileResponse(
        path=pangenome_full_path.as_posix(),
        filename=f"{pangenome.name}_id{pangenome.id}.h5",
    )


@router.get(
    "/pangenomes/{pangenome_id}/genomes",
    response_model=list[GenomePangenomeLinkPublic],
)
async def get_genomes_in_pangenome(
    pangenome_id: int,
    session: SessionDep,
    filter_genome: FilterGenome = Depends(),
    pagination_params: PaginationParams = Depends(),
):
    pangenome = pangenomes_crud.get_pangenome(session, pangenome_id)
    if not pangenome:
        raise HTTPException(status_code=404, detail="Pangenome not found")

    pangenome_with_genomes_stat = pangenomes_crud.get_genomes_in_pangenome(
        session,
        pangenome_id,
        filter_genome=filter_genome,
        pagination_params=pagination_params,
    )

    return pangenome_with_genomes_stat


@router.get(
    "/pangenomes/{pangenome_id}/{genome_id}",
    response_model=GenomePangenomeLinkPublic,
)
async def get_genome_in_pangenome(
    pangenome_id: int, genome_id: int, session: SessionDep
):

    pangenome = pangenomes_crud.get_pangenome(session, pangenome_id)
    if not pangenome:
        raise HTTPException(status_code=404, detail="Pangenome not found")

    pangenome_with_genomes_stat = pangenomes_crud.get_genome_in_pangenome(
        session, pangenome_id, genome_id
    )
    if not pangenome_with_genomes_stat:
        raise HTTPException(
            status_code=404,
            detail=f"Genome id={genome_id} not linked with pangenome id={pangenome_id}",
        )
    return pangenome_with_genomes_stat


@router.get(
    "/pangenomes/{pangenome_id}/{genome_id}/cgview_map",
    response_model=str,
    response_class=FileResponse,
)
async def get_genome_cgview_map(
    pangenome_id: int, genome_id: int, session: SessionDep, settings: SettingsDep
):

    pangenome = pangenomes_crud.get_pangenome(session, pangenome_id)

    if not pangenome:
        raise HTTPException(status_code=404, detail="Pangenome not found")

    pangenome_with_genomes_stat = pangenomes_crud.get_genome_in_pangenome(
        session, pangenome_id, genome_id
    )

    if not pangenome_with_genomes_stat:
        raise HTTPException(
            status_code=404,
            detail=f"Genome id={genome_id} not linked with pangenome id={pangenome_id}",
        )

    cgview_map_relative_path = (
        Path(pangenome.collection_release.pangenomes_directory)
        / pangenome.name
        / "proksee"
        / f"{pangenome_with_genomes_stat.genome.name}.json.gz"
    )

    cgview_map_full_path = settings.pangbank_data_dir / cgview_map_relative_path
    # cgview_map_relative_path

    if not cgview_map_full_path.exists():
        raise HTTPException(
            status_code=404,
            detail=f"cgview map '{cgview_map_relative_path}' does not exist",
        )

    return FileResponse(
        cgview_map_full_path,
        filename=f"{pangenome.name}_id{pangenome.id}_{pangenome_with_genomes_stat.genome.name}.json",
        media_type="application/json",
        headers={"Content-Encoding": "gzip", "Content-Type": "application/json"},
    )


@router.get("/pangenomes/count/", response_model=int)
async def get_pangenome_count(
    session: SessionDep, filter_params: FilterGenomeTaxonGenomePangenome = Depends()
):
    pangenomes = pangenomes_crud.get_pangenomes(session, filter_params)

    return len(pangenomes)

@router.get(
    "/pangenomes/{pangenome_id}/dbg/graph", response_model=str, response_class=FileResponse,
    include_in_schema=False
)
async def get_pangenome_dbg(
    pangenome_id: int, session: SessionDep, settings: SettingsDep
):
    pangenome = pangenomes_crud.get_pangenome(session, pangenome_id)

    if not pangenome:
        raise HTTPException(status_code=404, detail="Pangenome not found")

    dbg_relative_path = (
        Path(pangenome.collection_release.pangenomes_directory) / f"../metapang/pangenome_dbg/{pangenome.name}/{pangenome.name}.dbg"
    )
    dbg_full_path = settings.pangbank_data_dir / dbg_relative_path

    if not dbg_full_path.exists():
        raise HTTPException(
            status_code=404,
            detail=f"DBG file {dbg_relative_path} does not exists",
        )
    return FileResponse(
        path=dbg_full_path.as_posix(),
        filename=f"{pangenome.name}_id{pangenome.id}.dbg",
    )


def get_annotation(pangenome_id: int, suffix: str, session: SessionDep, settings: SettingsDep) -> FileResponse:
    pangenome = pangenomes_crud.get_pangenome(session, pangenome_id)

    if not pangenome:
        raise HTTPException(status_code=404, detail="Pangenome not found")

    dbg_annotations_relative_path = (
        Path(pangenome.collection_release.pangenomes_directory) / f"../metapang/pangenome_dbg/{pangenome.name}/{pangenome.name}{suffix}.row_diff_brwt.annodbg"
    )
    dbg_annotations_full_path = settings.pangbank_data_dir / dbg_annotations_relative_path

    if not dbg_annotations_full_path.exists():
        raise HTTPException(
            status_code=404,
            detail=f"DBG annotations file {dbg_annotations_relative_path} does not exists",
        )

    return FileResponse(
        path=dbg_annotations_full_path.as_posix(),
        filename=f"{pangenome.name}_id{pangenome.id}{suffix}.row_diff_brwt.annodbg",
    )

@router.get(
    "/pangenomes/{pangenome_id}/dbg/family_annotations", response_model=str, response_class=FileResponse,
    include_in_schema=False
)
async def get_pangenome_dbg_annotations(
    pangenome_id: int, session: SessionDep, settings: SettingsDep
):
    return get_annotation(pangenome_id, "", session, settings)

@router.get(
    "/pangenomes/{pangenome_id}/dbg/genome_annotations", response_model=str, response_class=FileResponse,
    include_in_schema=False
)
async def get_pangenome_dbg_annotations_genomes(
    pangenome_id: int, session: SessionDep, settings: SettingsDep
):
    return get_annotation(pangenome_id, "_genomes", session, settings)




from collections.abc import Generator
from typing import Annotated

try:
    from fastapi import Depends
except ImportError:
    raise ImportError(
        "FastAPI is required for API dependencies. "
        "Install it with: pip install pangbank-api[fastapi]"
    )

from sqlmodel import Session

from .database import engine
from .config import Settings, get_settings


def get_session() -> Generator[Session, None, None]:
    with Session(engine) as session:
        yield session


SessionDep = Annotated[Session, Depends(get_session)]
SettingsDep = Annotated[Settings, Depends(get_settings)]



from sqlmodel import SQLModel, create_engine

from .config import get_settings

settings = get_settings()
sqlite_url = f"sqlite:///{settings.pangbank_db_path}"

connect_args = {"check_same_thread": False}
engine = create_engine(sqlite_url, echo=False, connect_args=connect_args)


def create_db_and_tables():
    SQLModel.metadata.create_all(engine)



from pathlib import Path
from pangbank_api.models import (
    TaxonomySourceBase,
    GenomeSourceBase,
    Collection,
    CollectionRelease,
    MetadataSourceBase,
    MetadataBase,
)
from pydantic import BaseModel, Field


class GenomeSourceInput(GenomeSourceBase):
    file: Path


class TaxonomySourceInput(TaxonomySourceBase):
    file: Path


class GenomeMetadataSourceInput(MetadataSourceBase):
    file: Path
    genome_name_to_genome_metadata: dict[str, list[MetadataBase]] = Field(
        default_factory=dict
    )


class CollectionReleaseInput(BaseModel):
    collection: Collection
    release: CollectionRelease
    taxonomy: TaxonomySourceInput
    genome_sources: list[GenomeSourceInput] = Field(default_factory=list)
    genome_metadata_sources: list[GenomeMetadataSourceInput] = Field(
        default_factory=list
    )



import logging
from pathlib import Path

from typing import Optional

import typer
from sqlmodel import Session
from typing_extensions import Annotated

from pangbank_api.database import create_db_and_tables, engine
from pangbank_api.manage_db.collections import (
    add_pangenomes_to_db,
    create_collection_release,
    delete_collection_release,
    delete_full_collection,
    print_collections,
)
from pangbank_api.manage_db.genome_metadata import (
    add_genome_metadata_source_to_db,
    update_genome_pangenome_links_with_specific_metadata,
    app as genome_metadata_app,
    parse_metadata_table,
)
from pangbank_api.manage_db.genomes import add_genomes_to_db
from pangbank_api.manage_db.taxonomy import (
    add_taxon_to_db,
    link_genomes_and_taxa,
    parse_taxonomy_file,
    create_taxonomy_source,
)
from pangbank_api.manage_db.utils import (
    parse_collection_release_input_json,
    set_up_logging_config,
)
from pangbank_api.models import GenomeMetadataSource, MetadataBase

cli = typer.Typer(
    no_args_is_help=True,
    add_completion=False,
    short_help="Manage pangbank database.",
    context_settings={"help_option_names": ["-h", "--help"]},
)


cli.add_typer(
    genome_metadata_app, name="genome-metadata", short_help="Manage genome metadata."
)


@cli.command(no_args_is_help=True)
def add_collection_release(
    collection_release_json: Path = typer.Argument(
        ...,
        help="Path to the collection release input json file.",
        exists=True,
        dir_okay=True,
    ),
    pangbank_data_dir: Annotated[
        Path,
        typer.Option(
            envvar="PANGBANK_DATA_DIR",
            help="Path to the pangbank data directory.",
            exists=True,
        ),
    ] = Path("./"),
):
    set_up_logging_config()

    data_input = parse_collection_release_input_json(
        collection_release_json, pangbank_data_dir
    )

    collection_input = data_input.collection
    collection_release_input = data_input.release
    taxonomy_input = data_input.taxonomy
    genome_metadata_source_inputs = data_input.genome_metadata_sources

    taxonomy_file = taxonomy_input.file

    genome_sources = data_input.genome_sources

    pangenome_dir = pangbank_data_dir / collection_release_input.pangenomes_directory

    genome_name_to_lineage = parse_taxonomy_file(taxonomy_file)
    lineages = set(genome_name_to_lineage.values())

    metadata_source_and_genome_name_to_metadatas: list[
        tuple[GenomeMetadataSource, dict[str, list[MetadataBase]]]
    ] = []
    metadata_sources: list[GenomeMetadataSource] = []

    create_db_and_tables()

    with Session(engine) as session:

        # Parse metadata sources and metadata tables
        logging.info("Parsing metadata sources and metadata tables")
        for genome_metadata_source_input in genome_metadata_source_inputs:
            genome_metadata_file = genome_metadata_source_input.file
            logging.info(
                f"Parsing metadata sources and metadata tables {genome_metadata_file}"
            )
            genome_name_to_metadatas = {
                genome_name: genome_metadata_list
                for genome_name, genome_metadata_list in parse_metadata_table(
                    genome_metadata_file
                )
            }

            genome_metadata_source = GenomeMetadataSource.model_validate(
                genome_metadata_source_input, from_attributes=True
            )
            genome_metadata_source = add_genome_metadata_source_to_db(
                genome_metadata_source, session
            )

            metadata_source_and_genome_name_to_metadatas.append(
                (genome_metadata_source, genome_name_to_metadatas)
            )
            metadata_sources.append(genome_metadata_source)

        genome_name_to_genome = add_genomes_to_db(genome_sources, session)

        taxonomy_source = create_taxonomy_source(taxonomy_input, session=session)

        name_to_taxon_by_depth = add_taxon_to_db(
            taxonomy_source,
            lineages,
            session,
        )

        link_genomes_and_taxa(
            genome_name_to_genome,
            genome_name_to_lineage,
            name_to_taxon_by_depth,
            session,
        )

        collection_release = create_collection_release(
            collection_input=collection_input,
            collection_release_input=collection_release_input,
            taxonomy_source=taxonomy_source,
            genome_metadata_sources=metadata_sources,
            session=session,
        )

        add_pangenomes_to_db(
            pangenome_dir,
            genome_name_to_genome=genome_name_to_genome,
            collection_release=collection_release,
            session=session,
        )

        # add_metadata_to_genomes_of_the_release(
        #     collection_release,
        #     metadata_source_and_genome_name_to_metadatas,
        #     session=session,
        # )

        update_genome_pangenome_links_with_specific_metadata(
            collection_release,
            metadata_source_and_genome_name_to_metadatas,
            session=session,
        )


@cli.command()
def list_collections():
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
    )

    create_db_and_tables()

    print_collections()


@cli.command(no_args_is_help=True)
def delete_collection(
    collection_name: Annotated[
        str, typer.Argument(help="Name of the collection to delete.")
    ],
    release_version: Annotated[
        Optional[str],
        typer.Option(
            help="Specific release version to delete. If not provided, the entire collection will be deleted."
        ),
    ] = None,
):
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
    )

    create_db_and_tables()

    with Session(engine) as session:
        if release_version:
            delete_collection_release(session, collection_name, release_version)

        else:
            delete_full_collection(session, collection_name)


if __name__ == "__main__":
    cli()



import gzip
import logging
from typing import List

from rich.progress import track
from sqlmodel import Session, select

from pangbank_api.models import Genome, GenomeSource

from pangbank_api.manage_db.input_models import GenomeSourceInput


def create_genomes(
    genome_sources: list[GenomeSource],
    source_to_genomes: dict[str, list[str]],
    session: Session,
):
    """
    Assign genome sources to genomes and add new genomes to the database if they do not already exist.

    :param genome_sources: A list of genome sources to be processed.
    :param source_to_genomes: A dictionary mapping genome source names to lists of genome names.
    :param session: The database session.
    """

    genome_name_to_genomes: dict[str, Genome] = {}

    for genome_source in genome_sources:
        new_genomes: list[Genome] = []

        name_to_existing_genome = {
            genome.name: genome
            for genome in session.exec(
                select(Genome).where(Genome.genome_source == genome_source)
            ).all()
        }
        logging.info(
            f"Found {len(name_to_existing_genome)} genomes from '{genome_source.name}' in the database."
        )

        for genome_name in track(
            source_to_genomes[genome_source.name],
            description=f"Processing genomes from {genome_source.name}",
        ):
            if genome_name in name_to_existing_genome:
                genome = name_to_existing_genome[genome_name]
            else:
                genome = Genome(name=genome_name)
                new_genomes.append(genome)

            genome_name_to_genomes[genome_name] = genome

        if new_genomes:
            logging.info(f"Adding {len(new_genomes)} new genomes to the database.")
            session.add_all(new_genomes)

            for new_genome in new_genomes:
                new_genome.genome_source = genome_source

        else:
            logging.info(
                "No new genomes to add. All provided genomes are already present in the database."
            )
        session.refresh(genome_source)
        session.commit()

    return genome_name_to_genomes


def parse_genome_to_source_files(
    genome_source_inputs: list[GenomeSourceInput],
) -> dict[str, list[str]]:
    """
    Parse genome-to-source mapping from multiple source files.

    :param source_genomes_files: A list of file paths, each containing genome names associated with a source.
    :return: A dictionary mapping genome names to their respective source names.
    """

    source_to_genomes: dict[str, list[str]] = {}

    for genome_source_input in genome_source_inputs:
        open_func = gzip.open if genome_source_input.file.suffix == ".gz" else open

        with open_func(genome_source_input.file, "rt") as fl:
            source_to_genomes[genome_source_input.name] = [line.strip() for line in fl]

    return source_to_genomes


def create_genome_sources(
    genome_source_inputs: list[GenomeSourceInput], session: Session
):
    """
    Create genome sources from the provided inputs and add them to the database.

    :param genome_source_inputs: A list of genome source inputs.
    :param session: The database session.
    :return: A list of created or existing genome sources.
    """
    genome_sources: List[GenomeSource] = []

    for genome_source_input in genome_source_inputs:
        statement = select(GenomeSource).where(
            GenomeSource.name == genome_source_input.name
        )
        genome_source = session.exec(statement).first()

        if genome_source is None:
            logging.info(f"Creating a new GenomeSource: {genome_source_input.name}")
            genome_source = GenomeSource.model_validate(genome_source_input)
        else:
            logging.info(
                f"GenomeSource '{genome_source_input.name}' already exists in the database"
            )

        session.add(genome_source)

        genome_sources.append(genome_source)

    session.commit()

    for genome_source in genome_sources:
        session.refresh(genome_source)

    return genome_sources


def add_genomes_to_db(
    genome_source_inputs: list[GenomeSourceInput],
    session: Session,
) -> dict[str, Genome]:
    """ """
    logging.info("Starting to process genomes.")

    genome_sources = create_genome_sources(genome_source_inputs, session)

    source_to_genomes = parse_genome_to_source_files(genome_source_inputs)

    logging.info("Parsing genome source files completed.")

    genome_name_to_genomes = create_genomes(
        genome_sources=genome_sources,
        source_to_genomes=source_to_genomes,
        session=session,
    )
    logging.info("Adding genome to db completed.")

    return genome_name_to_genomes



import gzip
import logging
from pathlib import Path

import typer
from rich.progress import Progress, track
from sqlmodel import Session, select

from pangbank_api.models import (
    Genome,
    GenomeTaxonLink,
    Taxon,
    TaxonomySource,
)

from pangbank_api.manage_db.input_models import TaxonomySourceInput

app = typer.Typer(no_args_is_help=True)


def parse_taxonomy_file(taxonomy_file: Path) -> dict[str, tuple[str, ...]]:
    genome_to_lineage: dict[str, tuple[str, ...]] = {}

    proper_open = gzip.open if taxonomy_file.suffix == ".gz" else open

    with proper_open(taxonomy_file, "rt") as fl:  # 'rt' mode to read as text
        for line in fl:
            genome_name, taxonomy_str = line.strip().split("\t")
            genome_to_lineage[genome_name] = tuple(
                name.strip() for name in taxonomy_str.split(";")
            )

    return genome_to_lineage


def parse_ranks_str(ranks_str: str) -> list[str]:
    ranks = [rank.strip().title() for rank in ranks_str.split(";")]

    return ranks


def create_taxonomy_source(
    taxonomy_source_input: TaxonomySourceInput, session: Session
) -> TaxonomySource:
    # Check if taxonomy release already exists in DB
    statement = select(TaxonomySource).where(
        (TaxonomySource.name == taxonomy_source_input.name)
        & (TaxonomySource.version == taxonomy_source_input.version)
    )

    taxonomy_source = session.exec(statement).first()

    if taxonomy_source is None:
        logging.info(
            f"Taxonomy source '{taxonomy_source_input.name}' "
            f"(version {taxonomy_source_input.version}) not found in the database. Adding it now."
        )
        taxonomy_source = TaxonomySource.model_validate(taxonomy_source_input)

        session.add(taxonomy_source)

        session.commit()
        session.refresh(taxonomy_source)

    else:
        # the taxonomy release exists already
        # checking that given ranks are identical that ones attached to the taxonomy release

        if parse_ranks_str(taxonomy_source.ranks) != parse_ranks_str(
            taxonomy_source_input.ranks
        ):
            raise ValueError(
                f"Discrepancy in ranks for taxonomy_source {taxonomy_source}. "
                f"Existing ranks : {parse_ranks_str(taxonomy_source.ranks)} "
                f"vs given ranks {parse_ranks_str(taxonomy_source_input.ranks)}"
            )

        logging.info(
            f"taxonomy_source {taxonomy_source_input.name}:{taxonomy_source_input.version} already exist in DB"
        )

    return taxonomy_source


def get_common_taxa(taxa_A: list[Taxon], taxa_B: list[Taxon]) -> list[Taxon]:
    common_taxa: list[Taxon] = []
    for taxon in taxa_A:
        if taxon in taxa_B:
            common_taxa.append(taxon)

    return common_taxa


def get_taxa_by_depth(depth: int, taxonomy_source: TaxonomySource, session: Session):
    statement = select(Taxon).where(
        (Taxon.depth == depth) & (Taxon.taxonomy_source == taxonomy_source)
    )
    taxa = session.exec(statement).all()

    return taxa


def create_taxon_from_lineages(
    ranks: list[str],
    lineages: set[tuple[str, ...]],
    taxonomy_source: TaxonomySource,
    session: Session,
):
    taxon_names_by_depths: list[set[str]] = [set() for _ in range(len(ranks))]
    new_taxa: list[Taxon] = []
    taxa: list[Taxon] = []

    for lineage in lineages:
        for i, taxon_name in enumerate(lineage):
            taxon_names_by_depths[i].add(taxon_name)

    taxa_count = sum(len(taxon_set) for taxon_set in taxon_names_by_depths)

    name_to_taxon_by_depth: list[dict[str, Taxon]] = []

    with Progress() as progress:
        progress_task = progress.add_task("Creating taxa", total=taxa_count)

        for depth, taxon_set in enumerate(taxon_names_by_depths):
            taxon_name_to_taxon: dict[str, Taxon] = {}
            name_to_taxon_by_depth.append(taxon_name_to_taxon)

            name_to_taxon_at_depth = {
                taxon.name: taxon
                for taxon in get_taxa_by_depth(depth, taxonomy_source, session)
            }

            for taxon_name in taxon_set:
                try:
                    taxon = name_to_taxon_at_depth[taxon_name]
                except KeyError:
                    taxon = Taxon(
                        name=taxon_name,
                        rank=ranks[depth],
                        depth=depth,
                    )
                    new_taxa.append(taxon)

                taxa.append(taxon)
                taxon_name_to_taxon[taxon.name] = taxon

                progress.update(progress_task, advance=1)

    logging.info(f"Created {len(new_taxa)} new taxa out of {len(taxa)} total taxa.")
    session.add_all(new_taxa)

    for new_taxon in new_taxa:
        new_taxon.taxonomy_source = taxonomy_source

    session.refresh(taxonomy_source)
    session.commit()

    return name_to_taxon_by_depth


def link_genomes_and_taxa(
    genome_name_to_genome: dict[str, Genome],
    genome_name_to_lineage: dict[
        str,
        tuple[str, ...],
    ],
    name_to_taxon_by_depth: list[dict[str, Taxon]],
    session: Session,
):
    new_genome_taxon_links: list[GenomeTaxonLink] = []
    logging.info(f"Linking {len(genome_name_to_genome)} genomes.")

    linked_genome_count = 0
    unlinked_genomes_count = 0
    for genome_name, genome in track(
        genome_name_to_genome.items(), "Linking genomes to taxa"
    ):
        lineage = genome_name_to_lineage[genome_name]

        taxon_name = lineage[0]
        taxon = name_to_taxon_by_depth[0][taxon_name]

        existing_link = session.get(GenomeTaxonLink, (genome.id, taxon.id))
        if existing_link is None:
            unlinked_genomes_count += 1
            for depth, taxon_name in enumerate(lineage):
                taxon = name_to_taxon_by_depth[depth][taxon_name]

                new_genome_taxon_links.append(
                    GenomeTaxonLink(genome_id=genome.id, taxon_id=taxon.id)
                )
        else:
            linked_genome_count += 1

    logging.info(
        f"{linked_genome_count} genomes were already linked in the GenomeTaxonLink table."
    )
    logging.info(
        f"{unlinked_genomes_count} genomes were not linked in the GenomeTaxonLink table."
    )
    logging.info(
        f"Adding {len(new_genome_taxon_links)} new GenomeTaxonLink entries to the database."
    )
    session.add_all(new_genome_taxon_links)

    session.commit()


def add_taxon_to_db(
    taxonomy_source: TaxonomySource,
    lineages: set[tuple[str, ...]],
    session: Session,
):
    ranks = [rank.strip() for rank in taxonomy_source.ranks.split(";")]

    logging.info(
        f"Adding taxa from taxonomy source '{taxonomy_source.name}' to the database."
    )
    name_to_taxon_by_depth = create_taxon_from_lineages(
        ranks=ranks,
        lineages=lineages,
        taxonomy_source=taxonomy_source,
        session=session,
    )
    logging.info(
        f"Completed adding taxa from taxonomy source '{taxonomy_source.name}' to the database."
    )
    return name_to_taxon_by_depth



from sqlalchemy import func
from sqlmodel import Session, select

from pangbank_api.crud.common import (
    FilterGenomeTaxon,
    PaginationParams,
    get_taxonomies_from_taxa,
)
from pangbank_api.models import (
    Genome,
    GenomePublicWithTaxonomies,
    GenomeTaxonLink,
    Taxon,
    TaxonomyPublic,
    GenomeSourcePublic,
)


def get_genome_public(genome: Genome) -> GenomePublicWithTaxonomies:
    taxonomies = get_taxonomies_from_taxa(genome.taxa)
    genome_public = GenomePublicWithTaxonomies(
        **genome.model_dump(),
        taxonomies=[TaxonomyPublic(**taxonomy.model_dump()) for taxonomy in taxonomies],
        genome_source=GenomeSourcePublic(**genome.genome_source.model_dump()),
    )
    return genome_public


def get_genome_by_id(
    session: Session, genome_id: int
) -> GenomePublicWithTaxonomies | None:
    genome = session.get(Genome, genome_id)
    if genome is None:
        return None

    return get_genome_public(genome)


def get_genome_by_name(
    session: Session, genome_name: str
) -> GenomePublicWithTaxonomies | None:
    genome = session.exec(select(Genome).where(Genome.name == genome_name)).first()

    if genome is None:
        return None

    return get_genome_public(genome)


def get_genomes(
    session: Session,
    filter_params: FilterGenomeTaxon,
    pagination_params: PaginationParams | None,
) -> list[GenomePublicWithTaxonomies]:
    query = select(Genome).distinct()

    if filter_params.genome_name is not None:
        query = query.where(Genome.name == filter_params.genome_name)

    if filter_params.taxon_name is not None:
        # Apply offset and limit

        query = query.join(GenomeTaxonLink).join(Taxon)

        if filter_params.substring_taxon_match:
            query = query.where(
                func.lower(Taxon.name).like(f"%{filter_params.taxon_name.lower()}%")
            )
        else:
            # exact match
            query = query.where(Taxon.name == filter_params.taxon_name)

    if pagination_params:
        query = query.offset(pagination_params.offset).limit(pagination_params.limit)

    db_genomes = session.exec(query).all()

    public_genomes = [get_genome_public(genome) for genome in db_genomes]

    return public_genomes



from contextlib import asynccontextmanager

try:
    from fastapi import FastAPI
    from fastapi.middleware.cors import CORSMiddleware
except ImportError:
    raise ImportError(
        "FastAPI is required to run the API server. "
        "Install it with: pip install pangbank-api[fastapi]"
    )

from .database import create_db_and_tables
from .routers import collections, genomes, pangenomes
from .config import get_settings

@asynccontextmanager
async def lifespan(app: FastAPI):
    create_db_and_tables()
    yield

settings = get_settings()

app = FastAPI(
    lifespan=lifespan,
    title="PanGBank API",
    docs_url="/",
    description="API for managing collections pangenomes.",
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.pangbank_origins.split(";"),
    allow_credentials=True,
    allow_methods=["GET"],
    allow_headers=["*"],
)


app.include_router(collections.router)
app.include_router(genomes.router)
app.include_router(pangenomes.router)



from fastapi import APIRouter, Depends, HTTPException
from fastapi.responses import FileResponse

from pangbank_api.crud import collections as collections_crud
from pangbank_api.crud.common import FilterCollection, FilterRelease

from ..dependencies import SessionDep, SettingsDep
from ..models import CollectionPublicWithReleases


router = APIRouter(
    tags=["collections"],
    responses={404: {"description": "Not found"}},
)


@router.get("/collections/", response_model=list[CollectionPublicWithReleases])
def get_collections(session: SessionDep, filter_params: FilterCollection = Depends()):
    collections = collections_crud.get_collections(session, filter_params)

    return collections


@router.get("/collections/{collection_id}", response_model=CollectionPublicWithReleases)
def get_collection(
    collection_id: int, session: SessionDep, filter_release: FilterRelease = Depends()
):

    collection = collections_crud.get_collection(session, collection_id, filter_release)

    if not collection:
        raise HTTPException(
            status_code=404, detail=f"Collection with id={collection_id} not found"
        )

    return collection


@router.get(
    "/collections/{collection_id}/mash_sketch",
    response_model=str,
    response_class=FileResponse,
)
async def get_collection_mash_sketch(
    collection_id: int, session: SessionDep, settings: SettingsDep
):

    mash_sketch_file = collections_crud.get_collection_mash_sketch(
        session, collection_id
    )
    if not mash_sketch_file:
        raise HTTPException(
            status_code=404,
            detail=f"Mash sketch of collection with id={collection_id} not found",
        )
    mash_sketch_path = settings.pangbank_data_dir / mash_sketch_file

    if not mash_sketch_path.exists():
        raise HTTPException(
            status_code=404,
            detail=f"Pangenome file {mash_sketch_file.name} does not exists",
        )
    return FileResponse(path=mash_sketch_path.as_posix(), filename="mash_sketch.msh")

@router.get(
    "/collections/{collection_id}/index/info",
    response_model=str,
    response_class=FileResponse,
    include_in_schema=False
)
async def get_collection_index_info(
    collection_id: int, session: SessionDep, settings: SettingsDep
):
    index_directory = collections_crud.get_collection_index_directory(session, collection_id)

    if not index_directory:
        raise HTTPException(
            status_code=404,
            detail=f"Index info of collection with id={collection_id} not found",
        )

    index_info_file = settings.pangbank_data_dir / index_directory / "index_info.json"

    if not index_info_file.exists():
        raise HTTPException(
            status_code=404,
            detail=f"Index info file {index_info_file.name} does not exists",
        )

    return FileResponse(path=index_info_file, filename="index_info.json")

@router.get(
    "/collections/{collection_id}/index/pangenomes",
    response_model=str,
    response_class=FileResponse,
    include_in_schema=False
)
async def get_collection_index_pangenomes(
    collection_id: int, session: SessionDep, settings: SettingsDep
):
    index_directory = collections_crud.get_collection_index_directory(session, collection_id)

    if not index_directory:
        raise HTTPException(
            status_code=404,
            detail=f"Index pangenomes of collection with id={collection_id} not found",
        )

    index_pangenome_file = settings.pangbank_data_dir / index_directory / "pangenome_index.sbt.zip"

    if not index_pangenome_file.exists():
        raise HTTPException(
            status_code=404,
            detail=f"Index pangenome file {index_pangenome_file.name} does not exists",
        )

    return FileResponse(path=index_pangenome_file, filename="pangenome_index.sbt.zip")

@router.get(
    "/collections/{collection_id}/index/genomes",
    response_model=str,
    response_class=FileResponse,
    include_in_schema=False
)
async def get_collection_index_genomes(
    collection_id: int, session: SessionDep, settings: SettingsDep
):
    index_directory = collections_crud.get_collection_index_directory(session, collection_id)

    if not index_directory:
        raise HTTPException(
            status_code=404,
            detail=f"Index genomes of collection with id={collection_id} not found",
        )

    index_genome_file = settings.pangbank_data_dir / index_directory / "genome_index.sbt.zip"

    if not index_genome_file.exists():
        raise HTTPException(
            status_code=404,
            detail=f"Index genome file {index_genome_file.name} does not exists",
        )

    return FileResponse(path=index_genome_file, filename="genome_index.sbt.zip")



import csv
import gzip
import logging
from pathlib import Path
from typing import Iterator, List
from packaging.version import parse

# import sys
import yaml
from rich.console import Console
from rich.progress import track
from rich.table import Table
from sqlmodel import Session, select

from pangbank_api.database import engine
from pangbank_api.manage_db.taxonomy import get_common_taxa
from pangbank_api.models import (
    Collection,
    CollectionRelease,
    Genome,
    GenomeInPangenomeMetric,
    GenomePangenomeLink,
    Pangenome,
    PangenomeMetric,
    PangenomeTaxonLink,
    Taxon,
    GenomeMetadataSource,
    TaxonomySource,
)

# from pangbank_api.manage_db.genome_metadata import parse_metadata_table
from pangbank_api.manage_db.utils import compute_md5

logger = logging.getLogger(__name__)  # __name__ ensures uniqueness per module


def create_collection_release(
    collection_input: Collection,
    collection_release_input: CollectionRelease,
    taxonomy_source: TaxonomySource,
    genome_metadata_sources: list[GenomeMetadataSource],
    session: Session,
) -> CollectionRelease:
    """
    Create or retrieve a collection release from the database.

    This function:
    - Ensures the collection exists in the database, creating it if necessary.
    - Ensures the collection release exists in the database, creating it if necessary.
    - Validates version consistency between the input file and existing database records.
    """

    collection = Collection.model_validate(collection_input)

    statement = select(Collection).where((Collection.name == collection.name))

    collection_from_db = session.exec(statement).first()

    if collection_from_db is None:
        logger.info(f"Adding a new collection to DB: {collection.name}")
        session.add(collection)
        session.commit()
    else:
        collection = collection_from_db
        logger.info(f"Collection {collection.name} already exists in DB")

    collection_release = CollectionRelease.model_validate(collection_release_input)

    statement = (
        select(CollectionRelease)
        .join(Collection)
        .where(
            (Collection.name == collection.name)
            & (CollectionRelease.version == collection_release.version)
        )
    )

    collection_release_from_db = session.exec(statement).first()

    if collection_release_from_db is None:
        logger.info(
            f"Adding a new collection release to DB: {collection.name}:{collection_release.version}"
        )
        collection_release.collection = collection
        collection_release.taxonomy_source = taxonomy_source
        collection_release.genome_metadata_sources = genome_metadata_sources
        logger.info(f"New release is linked to taxonomy source {taxonomy_source.name}")
        logger.info(
            f"New release is linked to genome metadata sources {[source.name for source in genome_metadata_sources]}"
        )

        session.add(collection_release)
        session.commit()

    else:
        logger.info(
            f"Collection release {collection.name}:{collection_release.version} already exists in DB"
        )

        same_ppanggo_version = (
            collection_release.ppanggolin_version
            == collection_release_from_db.ppanggolin_version
        )
        same_pangbank_wf_version = (
            collection_release.pangbank_wf_version
            == collection_release_from_db.pangbank_wf_version
        )

        same_genome_metadata_sources = {
            source.name for source in genome_metadata_sources
        } == {
            source.name for source in collection_release_from_db.genome_metadata_sources
        }

        if (
            not same_ppanggo_version
            or not same_pangbank_wf_version
            or not same_genome_metadata_sources
        ):
            raise ValueError(
                f"For collection {collection.name} release {collection_release.version}:"
                "Not the same ppanggolin_version or pangbank_wf_version or genome_metadata_sources from input file and whats in the DB.. "
                f"ppanggolin version : {collection_release.ppanggolin_version} vs {collection_release.ppanggolin_version} "
                f"ppanggolin version : {collection_release.pangbank_wf_version} vs {collection_release.pangbank_wf_version} "
                f"genome_metadata_sources : {genome_metadata_sources} vs {collection_release_from_db.genome_metadata_sources}"
            )

        collection_release = collection_release_from_db

    set_latest_release_in_collection(collection, session)

    session.commit()
    session.refresh(collection)

    return collection_release


def set_latest_release_in_collection(collection: Collection, session: Session):
    """
    Set the latest collection release for a collection.
    """
    releases = sorted(collection.releases, key=lambda x: parse(x.version), reverse=True)

    # Mark the latest release
    if releases:
        releases[0].latest = True
        logger.info(
            f"Marked release {releases[0].version} as latest for collection {collection.name}"
        )
        for release in releases[1:]:
            release.latest = False
            session.add(release)


def parse_genomes_hash_file(genomes_md5sum_file: Path) -> dict[str, dict[str, str]]:
    """
    Parse a genome hash file to extract genome name-to-metadata mapping.

    :param genomes_md5sum_file: Path to the TSV file containing genome MD5 checksums.
    :return: A dictionary mapping genome names to their metadata, extracted from the file.

    The input file is expected to be a tab-separated values (TSV) file with at least
    a 'name' column for genome identifiers.
    """

    if not genomes_md5sum_file.exists():
        raise FileNotFoundError(
            f"Genome MD5 checksum file not found: {genomes_md5sum_file}"
        )
    proper_open = gzip.open if genomes_md5sum_file.suffix == ".gz" else open

    with proper_open(genomes_md5sum_file, "rt") as fl:
        genome_name_to_genome_info = {
            genome_info["name"]: genome_info
            for genome_info in csv.DictReader(fl, delimiter="\t")
        }

    return genome_name_to_genome_info


def parse_genome_metrics_file(tsv_file_path: Path) -> Iterator[GenomeInPangenomeMetric]:
    """
    Parse a TSV file containing genome data and yield GenomeInPangenomeMetric instances.

    :param tsv_file_path: Path to the TSV (or TSV.GZ) file containing genome metrics.
    :return: An iterator of GenomeInPangenomeMetric instances.

    The function:
    - Supports both uncompressed (.tsv) and gzip-compressed (.tsv.gz) files.
    - Skips comment lines (lines starting with `#`).
    - Converts column headers to lowercase before validation.
    - Raises a ValueError if a row fails to validate.
    """

    if not tsv_file_path.exists():
        raise FileNotFoundError(f"Genome metrics file not found: {tsv_file_path}")

    open_func = gzip.open if tsv_file_path.suffix == ".gz" else open

    with open_func(tsv_file_path, mode="rt", newline="", encoding="utf-8") as file:
        reader = csv.DictReader(filter(lambda row: row[0] != "#", file), delimiter="\t")

        for row in reader:
            try:
                genome_data = GenomeInPangenomeMetric.model_validate(
                    {key.lower(): value for key, value in row.items()}
                )
                yield genome_data

            except ValueError as e:
                raise ValueError(f"Error parsing row {row}: {e}") from e


def get_pangenome_metrics_from_genome_stats_summary_yaml(
    yaml_genome_stats_summary_path: Path,
):
    with open(yaml_genome_stats_summary_path, "r") as file:
        data = yaml.safe_load(file)

    pangenome_data = {
        "mean_completeness": data.get("Completeness", {}).get("mean"),
        "mean_contamination": data.get("Contamination", {}).get("mean"),
        "mean_fragmentation": data.get("Fragmentation", {}).get("mean"),
        "mean_exact_core_families_count_per_genome": data.get(
            "Exact_core_families", {}
        ).get("mean"),
        "mean_soft_core_families_count_per_genome": data.get(
            "Soft_core_families", {}
        ).get("mean"),
        "mean_persistent_families_count_per_genome": data.get(
            "Persistent_families", {}
        ).get("mean"),
        "mean_shell_families_count_per_genome": data.get("Shell_families", {}).get(
            "mean"
        ),
        "mean_cloud_families_count_per_genome": data.get("Cloud_families", {}).get(
            "mean"
        ),
    }

    return pangenome_data


def get_pangenome_metrics_from_info_yaml(yaml_info_path: Path):
    """ """

    with open(yaml_info_path, "r") as file:
        data = yaml.safe_load(file)

    # Initialize pangenome_data
    pangenome_data = {
        "gene_count": data.get("Content", {}).get("Genes"),
        "genome_count": data.get("Content", {}).get("Genomes"),
        "family_count": data.get("Content", {}).get("Families"),
        "edge_count": data.get("Content", {}).get("Edges"),
        # Other information
        "partition_count": data.get("Content", {}).get("Number_of_partitions"),
        "rgp_count": data.get("Content", {}).get("RGP"),
        "spot_count": data.get("Content", {}).get("Spots"),
        # Modules information
        "module_count": data.get("Content", {})
        .get("Modules", {})
        .get("Number_of_modules"),
        "family_in_module_count": data.get("Content", {})
        .get("Modules", {})
        .get("Families_in_Modules"),
        "all_genome_fluidity": data.get("Content", {})
        .get("Genomes_fluidity", {})
        .get("all"),
        "shell_genome_fluidity": data.get("Content", {})
        .get("Genomes_fluidity", {})
        .get("shell"),
        "cloud_genome_fluidity": data.get("Content", {})
        .get("Genomes_fluidity", {})
        .get("cloud"),
        "accessory_genome_fluidity": data.get("Content", {})
        .get("Genomes_fluidity", {})
        .get("accessory"),
    }

    # Loop through partitions to populate family counts and genome frequency metrics
    for partition in ["Persistent", "Shell", "Cloud"]:
        partition_key = partition.lower()

        pangenome_data[f"{partition_key}_family_count"] = (
            data.get("Content", {}).get(partition, {}).get("Family_count")
        )
        pangenome_data[f"{partition_key}_family_min_genome_frequency"] = (
            data.get("Content", {}).get(partition, {}).get("min_genomes_frequency")
        )
        pangenome_data[f"{partition_key}_family_max_genome_frequency"] = (
            data.get("Content", {}).get(partition, {}).get("max_genomes_frequency")
        )
        pangenome_data[f"{partition_key}_family_std_genome_frequency"] = (
            data.get("Content", {}).get(partition, {}).get("sd_genomes_frequency")
        )
        pangenome_data[f"{partition_key}_family_mean_genome_frequency"] = (
            data.get("Content", {}).get(partition, {}).get("mean_genomes_frequency")
        )
    return pangenome_data


def get_pangenome_metrics_from_info_files(
    yaml_info_path: Path, yaml_genome_stats_summary_path: Path
) -> PangenomeMetric:
    """ """

    pangenome_data = get_pangenome_metrics_from_info_yaml(yaml_info_path)
    pangenome_data.update(
        get_pangenome_metrics_from_genome_stats_summary_yaml(
            yaml_genome_stats_summary_path
        )
    )

    return PangenomeMetric.model_validate(pangenome_data)


def add_pangenomes_to_db(
    pangenome_main_dir: Path,
    collection_release: CollectionRelease,
    genome_name_to_genome: dict[str, Genome],
    session: Session,
) -> list[Pangenome]:
    """
    Parse the pangenome directory and load pangenome data into the database.

    :param pangenome_main_dir: The main directory containing pangenome subdirectories.
    :param collection_release: The collection release to associate the pangenomes with.
    :param session: SQLAlchemy session for database transactions.
    :return: A list of Pangenome objects.
    """
    pangenomes: List[Pangenome] = []

    pangenome_dirs = [
        pangenome_dir
        for pangenome_dir in pangenome_main_dir.iterdir()
        if pangenome_dir.is_dir()
    ]
    logger.info(f"Found {len(pangenome_dirs)} pangenome directories")

    existing_pangenomes = session.exec(
        select(Pangenome).where(Pangenome.collection_release == collection_release)
    ).all()

    logger.info(f"Found {len(existing_pangenomes)} existing pangenomes in the database")

    file_to_existing_pangenome = {
        pangenome.file_name: pangenome for pangenome in existing_pangenomes
    }

    new_pangenomes: List[Pangenome] = []

    for pangenome_dir in track(
        pangenome_dirs,
        description="Parsing pangenome directories",
    ):
        pangenome_file = pangenome_dir / "pangenome.h5"
        genomes_md5sum_file = (
            pangenome_dir / "genomes_md5sum.tsv.gz"
            if (pangenome_dir / "genomes_md5sum.tsv.gz").exists()
            else pangenome_dir / "genomes_md5sum.tsv"
        )
        pangenome_info_file = pangenome_dir / "info.yaml"
        yaml_genome_stats_summary_file = (
            pangenome_dir / "genomes_statistics_summary.yaml"
        )
        genomes_statistics_file = pangenome_dir / "genomes_statistics.tsv.gz"

        # genomes_metadata_dir = pangenome_dir / "metadata"

        pangenome_local_path = Path(pangenome_file.parent.name) / pangenome_file.name

        pangenome = file_to_existing_pangenome.get(
            pangenome_local_path.as_posix(), None
        )
        if pangenome is None:
            pangenome_file_md5sum = compute_md5(pangenome_file)

            pangenome_metric = get_pangenome_metrics_from_info_files(
                pangenome_info_file, yaml_genome_stats_summary_file
            )

            pangenome = Pangenome.model_validate(
                pangenome_metric,
                from_attributes=True,
                update={
                    "file_name": pangenome_local_path.as_posix(),
                    "collection_release": collection_release,
                    "name": pangenome_dir.name,
                    "file_md5sum": pangenome_file_md5sum,
                },
            )
            new_pangenomes.append(pangenome)
            genomes = link_pangenome_and_genomes(
                pangenome=pangenome,
                genome_name_to_genome=genome_name_to_genome,
                genomes_md5sum_file=genomes_md5sum_file,
                genomes_statistics_file=genomes_statistics_file,
                session=session,
            )

            link_pangenome_and_genome_taxa(
                pangenome, genomes, collection_release.taxonomy_source, session
            )

            # metadata_files = list(
            #     genomes_metadata_dir.glob("genomes_metadata_from_*.tsv*")
            # )

            # if genomes_metadata_dir.exists() and metadata_files:
            #     add_metadata_to_genome_pangenome_links(
            #         metadata_files, pangenome_genome_links, session
            #     )
        pangenomes.append(pangenome)

    session.add_all(new_pangenomes)
    session.commit()

    logger.info(f"Added {len(new_pangenomes)} new pangenomes to the database")

    return pangenomes


def extract_source_from_metadata_file(metadata_file: Path) -> str:
    """
    Extract the source name from a metadata file with a specific naming pattern.
    """

    prefix = "genomes_metadata_from_"
    valid_suffixes = {".tsv", ".tsv.gz"}

    filename = metadata_file.name
    suffix = "".join(metadata_file.suffixes)  # Handles single and multiple suffixes

    if not filename.startswith(prefix) or suffix not in valid_suffixes:
        raise ValueError(
            f"Invalid metadata file name '{filename}'. Expected format: '{prefix}<source>.tsv' or '{prefix}<source>.tsv.gz'"
        )

    source = metadata_file.stem  # Removes the last suffix (.tsv or .gz if present)

    if source.startswith(prefix):
        source = source[len(prefix) :]  # Remove prefix

    source = source.strip()

    if not source:
        raise ValueError(
            f"Metadata file name '{filename}' does not contain a valid source name."
        )

    return source


def link_pangenome_and_genome_taxa(
    pangenome: Pangenome,
    genomes: list[Genome],
    taxonomy_source: TaxonomySource,
    session: Session,
):
    pangenome_taxa: List[Taxon] = []

    for genome in genomes:
        genome_taxa = [
            taxon for taxon in genome.taxa if taxon.taxonomy_source == taxonomy_source
        ]
        if not pangenome_taxa:
            pangenome_taxa = genome_taxa
        else:
            pangenome_taxa = get_common_taxa(genome_taxa, pangenome_taxa)
    pangenome_taxon_links = [
        PangenomeTaxonLink(pangenome_id=pangenome.id, taxon_id=taxon.id)
        for taxon in pangenome_taxa
    ]
    session.add_all(pangenome_taxon_links)


def link_pangenome_and_genomes(
    pangenome: Pangenome,
    genome_name_to_genome: dict[str, Genome],
    genomes_md5sum_file: Path,
    genomes_statistics_file: Path,
    session: Session,
):
    genome_name_to_md5sum_info = parse_genomes_hash_file(genomes_md5sum_file)
    pangenome_genome_links: List[GenomePangenomeLink] = []

    # pangenome_taxa: List[Taxon] = []
    genomes: List[Genome] = []
    for genome_metric in parse_genome_metrics_file(genomes_statistics_file):
        genome = genome_name_to_genome[genome_metric.genome_name]

        genome_file_info = genome_name_to_md5sum_info[genome_metric.genome_name]

        pangenome_genome_link = GenomePangenomeLink.model_validate(
            genome_metric,
            from_attributes=True,
            update={
                "genome": genome,
                "pangenome": pangenome,
                "genome_file_md5sum": genome_file_info["md5_sum"],
                "genome_file_name": genome_file_info["file_name"],
            },
        )
        pangenome_genome_links.append(pangenome_genome_link)
        genomes.append(genome)
        # pangenome_taxa = get_common_taxa(genome.taxa, pangenome_taxa)

    # session.add(pangenome)
    session.add_all(pangenome_genome_links)

    return genomes
    # session.commit()

    # pangenome_taxon_links = [
    #     PangenomeTaxonLink(pangenome_id=pangenome.id, taxon_id=taxon.id)
    #     for taxon in pangenome_taxa
    # ]
    # session.add_all(pangenome_taxon_links)
    # link_pangenome_and_taxa(pangenome, pangenome_taxa, session)


def delete_full_collection(session: Session, collection_name: str) -> None:
    """
    Deletes a collection from the database if it exists.

    :param session: Database session used for querying and deleting the collection.
    :param collection_name: Name of the collection to delete.
    """
    # Query the database to find the collection with the specified name
    statement = select(Collection).where(Collection.name == collection_name)
    collection_from_db = session.exec(statement).first()

    if collection_from_db is None:
        raise ValueError(f"Collection '{collection_name}' not found in the database.")
    else:
        logger.info(f"Deleting collection '{collection_name}' from the database.")
        session.delete(collection_from_db)
        session.commit()


def delete_collection_release(
    session: Session, collection_name: str, release_version: str
) -> None:
    """
    Deletes a specific collection release from the database if it exists.

    :param session: Database session used for querying and deleting the collection release.
    :param collection_name: Name of the collection containing the release.
    :param release_version: Version of the collection release to delete.
    """
    # Query the database to find the collection release with the specified name and version
    statement = (
        select(CollectionRelease)
        .join(Collection)
        .where(
            (Collection.name == collection_name)
            & (CollectionRelease.version == release_version)
        )
    )
    collection_release_from_db = session.exec(statement).first()

    if collection_release_from_db is None:
        raise ValueError(
            f"Collection release '{collection_name}' (version: {release_version}) not found in the database."
        )
    else:
        logger.info(
            f"Deleting collection release '{collection_name}' (version: {release_version}) from the database."
        )
        session.delete(collection_release_from_db)
        session.commit()


def print_collections():
    """ """
    console = Console()
    with Session(engine) as session:
        statement = select(Collection)

        results = session.exec(statement).all()

        if not results:
            console.print("[bold red]No collections found in the database.[/bold red]")
            return

        for collection in results:
            table = Table(
                title=f"Collections {collection.name}",
                caption=collection.description,
                show_header=True,
                header_style="bold magenta",
            )
            table.add_column("Release", style="bold cyan")
            table.add_column("Pangenomes", justify="right")
            table.add_column("Note", style="dim")

            for collection_release in collection.releases:
                table.add_row(
                    str(collection_release.version),
                    str(len(collection_release.pangenomes)),
                    collection_release.release_note,
                )

            console.print(table)
